import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Function
from torch.nn.modules.utils import _pair
from torch.nn import init
import unoptimized_cuda_kernel
import math
import numpy as np
import time

class _UnoptimizedConvNd(nn.Module):

    __constants__ = ['stride', 'padding', 'dilation', 'groups', 'bias', 'padding_mode']

    def __init__(self, in_channels, out_channels, kernel_size, stride,
                 padding, dilation, transposed, output_padding,
                 groups, bias, padding_mode):
        super(_UnoptimizedConvNd, self).__init__()
        if in_channels % groups != 0:
            raise ValueError('in_channels must be divisible by groups')
        if out_channels % groups != 0:
            raise ValueError('out_channels must be divisible by groups')
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.dilation = dilation
        self.transposed = transposed
        self.output_padding = output_padding
        self.groups = groups
        self.padding_mode = padding_mode

        if transposed:
            self.weight = nn.Parameter(torch.Tensor(
                in_channels, out_channels // groups, *kernel_size))
        else:
            self.weight = nn.Parameter(torch.Tensor(
                out_channels, in_channels // groups, *kernel_size))
        if bias:
            self.bias = nn.Parameter(torch.Tensor(out_channels))
        else:
            self.register_parameter('bias', None)
        self.reset_parameters()

    def reset_parameters(self):
        init.kaiming_uniform_(self.weight, a=math.sqrt(5))

        if self.bias is not None:
            fan_in, _ = init._calculate_fan_in_and_fan_out(self.weight)
            bound = 1 / math.sqrt(fan_in)
            init.uniform_(self.bias, -bound, bound)

    def extra_repr(self):
        s = ('{in_channels}, {out_channels}, kernel_size={kernel_size}'
             ', stride={stride}')
        if self.padding != (0,) * len(self.padding):
            s += ', padding={padding}'
        if self.dilation != (1,) * len(self.dilation):
            s += ', dilation={dilation}'
        if self.output_padding != (0,) * len(self.output_padding):
            s += ', output_padding={output_padding}'
        if self.groups != 1:
            s += ', groups={groups}'
        if self.bias is None:
            s += ', bias=False'
        return s.format(**self.__dict__)

class UnoptimizedConv2d(_UnoptimizedConvNd):
    def __init__(self, in_channels, out_channels, kernel_size, stride=1,
                 padding=0, dilation=1, groups=1,
                 bias=True, padding_mode='zeros'):
        kernel_size = _pair(kernel_size)
        stride = _pair(stride)
        padding = _pair(padding)
        dilation = _pair(dilation)
        super(UnoptimizedConv2d, self).__init__(
            in_channels, out_channels, kernel_size, stride, padding, dilation,
            False, _pair(0), groups, bias, padding_mode)

    #@weak_script_method
    def forward(self, input):
        # start_time = time.time()
        if self.padding_mode == 'circular':
            print('circular')
        if len(self.padding) == 2:
            padding = (self.padding[0],self.padding[0],self.padding[1],self.padding[1])
        else:
            padding = self.padding
        input = F.pad(input = input, pad = padding, mode = 'constant', value = 0)
        if len(self.stride) == 1:
            strides_h = self.stride[0]
            strides_w = self.stride[0]
        else: 
            strides_h = self.stride[0]
            strides_w = self.stride[1]
        out_height = int((input.size(2) - self.weight.size(2)) / strides_h +1)
        out_width = int((input.size(3) - self.weight.size(3)) / strides_w +1)
        out = torch.zeros([input.size(0), self.weight.size(0), out_height, out_width], dtype=torch.float, device=torch.device('cuda:0'))

        if self.bias is not None:
            unoptimized_cuda_kernel.UNOPTIMIZED_CONV(input, self.weight, self.bias, out, self.stride, self.padding )
        else:
            temp = torch.zeros([self.weight.size(0)], dtype=torch.float, device=torch.device('cuda:0'))
            unoptimized_cuda_kernel.UNOPTIMIZED_CONV(input, self.weight, temp, out, self.stride, self.padding )
        # end_time = time.time()
        # print("Conv Time:", end_time - start_time )
        return out
